import datetime

import dash
import numpy as np
import pandas as pd
from dash import Dash, Input, Output, callback, dcc, html

from tradingo.api import Tradingo
from tradingo.settings import TradingoConfig

config = TradingoConfig.from_env().to_env()

ARCTIC_URL = config.arctic_uri

pd.options.plotting.backend = "plotly"


app = Dash()

app.layout = html.Div(
    [
        dcc.Location(id="url", refresh=False),
        html.Div(
            [
                html.H1(children="Tradingo Monitor", style={"textAlign": "center"}),
                dcc.DatePickerRange(
                    start_date=None,
                    end_date=None,
                    id="date-selection",
                ),
                dcc.Dropdown(
                    options=[],
                    placeholder="Universe",
                    value="im-multi-asset",
                    id="universe-selection",
                    multi=False,
                ),
                dcc.Dropdown(
                    value=None,
                    options=[],
                    placeholder="Symbols",
                    id="asset-selection",
                    multi=True,
                ),
                dcc.Dropdown(
                    value=None,
                    options=[],
                    placeholder="Portfolio",
                    id="portfolio-selection",
                    multi=False,
                ),
            ]
        ),
        html.Div(
            [
                dcc.Graph(id="day-to-date"),
                dcc.Graph(id="week-to-date"),
                dcc.Graph(id="month-to-date"),
                dcc.Graph(id="year-to-date"),
                dcc.Graph(id="z_score"),
                dcc.Graph(id="net_position"),
                dcc.Graph(id="raw_position"),
                dcc.Graph(id="unrealised_pnl"),
                dcc.Graph(id="total_pnl"),
                dcc.Graph(id="instrument_total_pnl"),
                dcc.Graph(id="gross_margin"),
                dcc.Graph(id="margin"),
                dcc.Graph(id="short_vol"),
                dcc.Graph(id="long_vol"),
            ]
        ),
    ]
)


@callback(
    Output("universe-selection", "options"),
    Input("url", "pathname"),
)
def set_universe_options(_):
    api = Tradingo(
        uri=ARCTIC_URL,
    )
    return api.instruments._library.list_symbols()


@callback(
    Output("asset-selection", "options"),
    Input("universe-selection", "value"),
)
def set_asset_options(universe):
    if not universe:
        return dash.no_update
    api = Tradingo(
        uri=ARCTIC_URL,
    )
    return api.instruments[universe]().index.to_list()


@callback(
    Output("portfolio-selection", "options"),
    Input("universe-selection", "value"),
)
def set_portfolio_options(universe):
    if not universe:
        return dash.no_update
    api = Tradingo(
        uri=ARCTIC_URL,
    )
    return api.portfolio.list()


@callback(
    Output("portfolio-selection", "value"),
    Input("portfolio-selection", "options"),
)
def set_portfolio_value(options):
    if not options:
        return dash.no_update
    return options[0]


@callback(
    (
        Output("z_score", "figure"),
        Output("net_position", "figure"),
        Output("raw_position", "figure"),
        Output("unrealised_pnl", "figure"),
        Output("total_pnl", "figure"),
        Output("instrument_total_pnl", "figure"),
        Output("margin", "figure"),
        Output("gross_margin", "figure"),
        Output("short_vol", "figure"),
        Output("long_vol", "figure"),
        Output("day-to-date", "figure"),
        Output("week-to-date", "figure"),
        Output("month-to-date", "figure"),
        Output("year-to-date", "figure"),
    ),
    (
        Input("asset-selection", "value"),
        Input("date-selection", "start_date"),
        Input("date-selection", "end_date"),
        Input("universe-selection", "value"),
        Input("portfolio-selection", "value"),
    ),
)
def update_graph(
    assets,
    start_date,
    end_date,
    universe,
    portfolio,
):
    if not universe or not portfolio:
        return dash.no_update

    api = Tradingo(
        uri=ARCTIC_URL,
    )

    def range_breaks(fig):
        fig.update_xaxes(
            rangebreaks=[
                dict(bounds=[22, 13], pattern="hour"),
                dict(bounds=["sat", "mon"], pattern="day of week"),
            ]
        )
        return fig

    assets = assets or None
    date = (
        pd.Timestamp(start_date or (pd.Timestamp.now() - pd.offsets.BDay(0)))
        .normalize()
        .tz_localize("utc")
    )
    end = (
        pd.Timestamp(
            end_date or (pd.Timestamp.now().normalize() + pd.Timedelta(hours=24))
        )
        .normalize()
        .tz_localize("utc")
    )

    z_score = api.signals.intraday_momentum.z_score[universe](
        columns=assets,
        date_range=(date, end),
    )

    short_vol = api.signals.intraday_momentum.short_vol[universe](
        date_range=((end or pd.Timestamp.now()) - pd.offsets.BDay(30), end),
        columns=assets,
    ).resample("B").last() * np.sqrt(252)
    long_vol = api.signals.intraday_momentum.long_vol[universe](
        date_range=((end or pd.Timestamp.now()) - pd.offsets.BDay(30), end),
        columns=assets,
    ).resample("B").last() * np.sqrt(252)

    start = z_score.index[0] if len(z_score.index) else date
    end = z_score.index[-1] if len(z_score.index) else date

    unrealised_pnl = (
        api.backtest[portfolio]
        .instrument.unrealised_pnl(
            columns=assets,
            date_range=(start, end),
        )
        .ffill()
        .fillna(0.0)
        .diff()
        .cumsum()
        .fillna(0.0)
    )

    net_position = api.backtest[portfolio].instrument.net_position(
        columns=assets,
        date_range=(start, end),
    )
    raw_position = api.portfolio[portfolio].raw.position(
        columns=assets,
        date_range=(start, end),
    )
    margin = (
        api.backtest[portfolio]
        .instrument.net_exposure(
            date_range=(start, end),
            columns=assets,
        )
        .abs()
        * 0.05
    )
    gross_margin = (
        api.backtest[portfolio]
        .portfolio(
            date_range=(start, end),
        )
        .gross_exposure.abs()
        * 0.05
    )
    total_pnl = (
        api.backtest[portfolio]
        .portfolio(
            date_range=(
                start,
                end,
            ),
        )
        .total_pnl.ffill()
        .fillna(0.0)
        .diff()
        .cumsum()
        .fillna(0.0)
    )
    instrument_total_pnl = (
        api.backtest[portfolio]
        .instrument.total_pnl(
            date_range=(
                start - pd.offsets.YearBegin(1),
                end,
            ),
        )
        .resample(pd.offsets.BDay(1))
        .last()
        .ffill()
        .fillna(0.0)
        .diff()
        .cumsum()
        .fillna(0.0)
    )

    returns = api.backtest[portfolio].portfolio(
        date_range=(
            end - pd.offsets.BDay(252),
            end,
        ),
    )
    one_year = (
        returns.resample(pd.offsets.BDay(1)).last().diff().fillna(0).cumsum().total_pnl
    )
    one_month = (
        returns.loc[end - pd.offsets.Day(30) : end]
        .resample(pd.offsets.BDay(1))
        .last()
        .diff()
        .fillna(0)
        .cumsum()
        .total_pnl
    )
    one_week = (
        returns.loc[end - pd.offsets.Week(1) : end]
        .resample(pd.offsets.Hour(4))
        .last()
        .diff()
        .fillna(0)
        .cumsum()
        .total_pnl
    )
    one_day = (
        returns[
            (returns.index.time < datetime.time.fromisoformat("21:30"))
            & (returns.index.time > datetime.time.fromisoformat("09:30"))
        ]
        .loc[end - pd.offsets.BDay(1) : end]
        .diff()
        .fillna(0)
        .cumsum()
        .total_pnl
    )

    return (
        range_breaks(z_score.plot(title="z_score")),
        range_breaks(net_position.plot(title="net_position")),
        range_breaks(raw_position.plot(title="raw_position")),
        range_breaks(unrealised_pnl.plot(title="unrealised_pnl")),
        range_breaks(total_pnl.plot(title="total_pnl")),
        instrument_total_pnl.plot(title="instrument_total_pnl"),
        range_breaks(
            (gross_margin - unrealised_pnl.sum(axis=1)).plot(title="total_margin")
        ),
        range_breaks(margin.plot(title="margin")),
        # range_breaks(short_vol.plot(title="short_vol")),
        short_vol.plot(title="short_vol"),
        long_vol.plot(title="long_vol"),
        one_day.plot(title="DTD"),
        one_week.plot(title="WTD"),
        one_month.plot(title="MTD"),
        one_year.plot(title="YTD"),
    )


server = app.server


if __name__ == "__main__":
    app.run(
        host="0.0.0.0",
        debug=True,
    )



from __future__ import annotations

import functools
import inspect
import logging
from collections import defaultdict
from typing import DefaultDict, NamedTuple, Optional
from urllib.parse import parse_qsl, urlparse

import arcticdb as adb
import pandas as pd
from arcticdb.exceptions import NoSuchVersionException
from arcticdb_ext.exceptions import InternalException

logger = logging.getLogger(__name__)


class SymbolParseError(Exception):
    """raised when cant parse symbol"""


def add_params(function, *args):
    origsig = inspect.signature(function)
    orig_params = list(origsig.parameters.values())
    for arg in args:
        if arg in origsig.parameters:
            continue
        else:
            orig_params.insert(
                0,
                inspect.Parameter(
                    arg,
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                ),
            )

    function.__signature__ = origsig.replace(parameters=orig_params)


class Symbol(NamedTuple):
    library: str
    symbol: str
    kwargs: dict

    @classmethod
    def parse(
        cls,
        base: str,
        kwargs: dict,
        symbol_prefix: str = "",
        symbol_postfix: str = "",
    ) -> Symbol:
        """
        Parse a symbol string and return a Symbol object.
        """

        string_kwargs = {k: str(v) for k, v in kwargs.items()}

        try:
            symbol = base.format(**string_kwargs)
            parsed_symbol = urlparse(symbol)
            try:
                lib, sym = parsed_symbol.path.split("/")
            except ValueError as ex:
                raise SymbolParseError(f"symbol {symbol} is invalid.") from ex
            kwargs = dict(parse_qsl(parsed_symbol.query))
            symbol_prefix = symbol_prefix.format(**string_kwargs)
            symbol_postfix = symbol_postfix.format(**string_kwargs)
        except KeyError as ex:
            raise SymbolParseError(
                f"Missing parameter: {ex.args[0]},"
                f" {symbol_prefix=}, {symbol_postfix=}, {base=}, {string_kwargs=}"
            )

        for key, value in kwargs.items():
            if key == "as_of":
                try:
                    kwargs[key] = int(value)
                except TypeError:
                    continue
            if key == "columns":
                kwargs[key] = list(kwargs[key].split(","))

        return cls(lib, symbol_prefix + sym + symbol_postfix, kwargs)


def lib_provider(**libs):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(*args, arctic: adb.Arctic, **kwargs):
            libs_ = {
                k: arctic.get_library(kwargs.get(k, v), create_if_missing=True)
                for k, v in libs.items()
            }
            kwargs.update(libs_)
            return envoke_symbology_function(
                func,
                args,
                kwargs,
                arctic,
            )

        setattr(wrapper, "lib_provided", None)

        add_params(wrapper, "arctic")

        return wrapper

    return decorator


def symbol_provider(
    symbol_prefix="",
    no_date=False,
    **symbols,
):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(
            *args,
            arctic: adb.Arctic,
            start_date=None,
            end_date=None,
            raise_if_missing: bool = True,
            **kwargs,
        ):
            orig_symbol_data = {}
            requested_symbols = symbols.copy()

            for symbol in symbols:
                if symbol in kwargs and isinstance(
                    kwargs[symbol], (pd.DataFrame, pd.Series)
                ):
                    orig_symbol_data[symbol] = kwargs.pop(symbol)
                if symbol in kwargs and isinstance(kwargs[symbol], str):
                    requested_symbols[symbol] = kwargs[symbol]
                if symbol in kwargs and kwargs[symbol] is None:
                    requested_symbols.pop(symbol)

            def get_symbol_data(v: str | list[str], with_no_date=False) -> pd.DataFrame:
                if isinstance(v, dict):
                    return {
                        key: get_symbol_data(item, with_no_date=with_no_date)
                        for key, item in v.items()
                    }
                if isinstance(v, list):
                    multidata = pd.concat(
                        (
                            get_symbol_data(item, with_no_date=with_no_date)
                            for item in v
                        ),
                        axis=1,
                        keys=v,
                    )
                    columns = multidata.columns.get_level_values(1).drop_duplicates(
                        keep="first"
                    )
                    multidata = (
                        multidata.transpose().groupby(level=1).last().transpose()
                    )
                    return multidata[columns]
                symbol = Symbol.parse(v, kwargs, symbol_prefix=symbol_prefix)
                try:
                    data = (
                        arctic.get_library(
                            symbol.library,
                            create_if_missing=True,
                        )
                        .read(
                            symbol.symbol,
                            date_range=(
                                None
                                if with_no_date
                                else (
                                    pd.Timestamp(start_date) if start_date else None,
                                    pd.Timestamp(end_date) if end_date else None,
                                )
                            ),
                            **symbol.kwargs,
                        )
                        .data
                    )
                    if (
                        not with_no_date
                        and isinstance(data.index, pd.DatetimeIndex)
                        and not data.index.tz
                    ):
                        data.index = data.index.tz_localize("UTC")
                    return data
                except InternalException as ex:
                    if "The data for this symbol is pickled" in ex.args[0]:
                        return get_symbol_data(v, with_no_date=True)
                    raise ex
                except NoSuchVersionException as ex:
                    if not raise_if_missing:
                        return None
                    raise ex

            symbols_data = {
                k: get_symbol_data(v, with_no_date=no_date) if v is not None else v
                for k, v in requested_symbols.items()
                if k not in orig_symbol_data
            }
            kwargs.update(symbols_data)
            kwargs.update(orig_symbol_data)

            logger.info("Providing %s symbols from %s", symbols_data.keys(), arctic)

            return envoke_symbology_function(
                func,
                args,
                kwargs,
                arctic=arctic,
                start_date=start_date,
                end_date=end_date,
            )

        setattr(wrapper, "is_provided", None)
        add_params(wrapper, "arctic", "start_date", "end_date")

        return wrapper

    return decorator


def envoke_symbology_function(
    function,
    args,
    kwargs,
    arctic,
    start_date=None,
    end_date=None,
):
    sig = inspect.signature(function)

    if "start_date" in sig.parameters:
        kwargs.setdefault("start_date", start_date)
    if "end_date" in sig.parameters:
        kwargs.setdefault("end_date", end_date)
    if "arctic" in sig.parameters:
        kwargs.setdefault("arctic", arctic)

    return function(*args, **kwargs)


def symbol_publisher(
    *symbols,
    symbol_prefix="",
    symbol_postfix="",
    astype=None,
    template=None,
    library_options=None,
    write_pickle=False,
):
    def decorator(func):
        @functools.wraps(func)
        def wrapper(
            *args,
            arctic: adb.Arctic,
            dry_run=True,
            snapshot: Optional[str] = None,
            clean: bool = False,
            **kwargs,
        ):
            out = envoke_symbology_function(func, args, kwargs, arctic)
            if not isinstance(out, (tuple, list)):
                out = (out,)
            logger.info("Publishing %s to %s", symbols or template, arctic)

            if template:
                out, symbols_ = tuple(zip(*out))
                formatted_symbols = tuple(template.format(*s) for s in symbols_)
            else:
                formatted_symbols = symbols

            libraries: DefaultDict[str, dict] = defaultdict(dict)

            for data, symbol in zip(out, formatted_symbols, strict=True):
                if data.empty:
                    continue

                if astype:
                    data = data.astype(
                        astype
                        if not isinstance(astype, dict)
                        else {k: v for k, v in astype.items() if k in data.columns}
                    )

                if not dry_run:
                    parsed_symbol = Symbol.parse(
                        symbol,
                        kwargs,
                        symbol_prefix=symbol_prefix,
                        symbol_postfix=symbol_postfix,
                    )
                    logger.info(
                        "writing symbol=%s rows=%s",
                        parsed_symbol,
                        len(data.index),
                    )

                    lib_name, sym, params = parsed_symbol
                    lib = arctic.get_library(
                        lib_name,
                        create_if_missing=True,
                        library_options=library_options,
                    )
                    if isinstance(data.index, pd.DatetimeIndex):
                        if clean:
                            lib.delete(sym)
                        result = lib.update(
                            sym,
                            data,
                            upsert=True,
                            date_range=(data.index[0], data.index[-1]),
                            **params,
                        )
                    elif write_pickle:
                        result = lib.write_pickle(sym, data, **params)

                    else:
                        result = lib.write(sym, data, **params)

                    libraries[lib.name][result.symbol] = result.version

            if not dry_run and snapshot:
                for lib_name, versions in libraries.items():
                    logging.info(
                        "Snapshotting %s for %s %s", lib_name, snapshot, versions
                    )
                    lib = arctic.get_library(lib_name)
                    if snapshot in lib.list_snapshots():
                        lib.delete_snapshot(snapshot)
                    lib.snapshot(snapshot_name=snapshot, versions=versions)

            if dry_run:
                return pd.concat(out, keys=formatted_symbols, axis=1)

            return None

        setattr(wrapper, "is_publisher", None)
        add_params(wrapper, "arctic", "dry_run", "snapshot", "clean")
        return wrapper

    return decorator



from arcticdb import Arctic

library = "prices"
symbol_regex = "^ig-trading"

# source_uri = "lmdb:///home/rory/dev/airflow/test/arctic.db"
target_uri = (
    "s3://s3.us-east-1.amazonaws.com:tradingo-store?aws_auth=true&path_prefix=prod"
)
source_uri = "lmdb:///home/rory/dev/tradingo-plat/data/prod/tradingo.db"


def sync_symbols(
    library,
    symbol_regex,
    source_uri,
    target_uri,
):
    source_a = Arctic(source_uri)
    target_a = Arctic(target_uri)

    source_lib = source_a.get_library(library)
    target_lib = target_a.get_library(library, create_if_missing=True)

    for sym in source_lib.list_symbols(regex=symbol_regex):
        data = source_lib.read(sym)
        target_lib.write(sym, data.data)



"""IG data accessors"""

import logging
from typing import Optional

import dateutil.tz
import numpy as np
import pandas as pd
from arcticdb.version_store.library import Library
from tenacity import Retrying, retry_if_exception_type, wait_exponential
from trading_ig.rest import ApiExceededException, IGService

from tradingo import symbols
from tradingo.settings import IGTradingConfig

logger = logging.getLogger(__name__)


def get_ig_service(
    username: Optional[str] = None,
    password: Optional[str] = None,
    api_key: Optional[str] = None,
    acc_type: Optional[str] = None,
) -> IGService:
    config = IGTradingConfig.from_env()

    retryer = Retrying(
        wait=wait_exponential(),
        retry=retry_if_exception_type(ApiExceededException),
    )

    service = IGService(
        username=username or config.username,
        password=password or config.password,
        api_key=api_key or config.api_key,
        acc_type=acc_type or config.acc_type,
        use_rate_limiter=True,
        retryer=retryer,
    )

    service.create_session()
    return service


def sample_instrument(
    epic: str,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    interval: str,
    wait: int = 0,
    service: Optional[IGService] = None,
):
    service = service or get_ig_service()
    try:
        result = (
            service.fetch_historical_prices_by_epic(
                epic,
                end_date=pd.Timestamp(end_date)
                .tz_convert(dateutil.tz.tzlocal())
                .tz_localize(None)
                .isoformat(),
                start_date=(pd.Timestamp(start_date) + pd.Timedelta(seconds=1))
                .tz_convert(dateutil.tz.tzlocal())
                .tz_localize(None)
                .isoformat(),
                resolution=interval,
                wait=wait,
            )["prices"]
            .tz_localize(dateutil.tz.tzlocal())
            .tz_convert("utc")
        )

    except Exception as ex:
        if ex.args and (
            ex.args[0] == "Historical price data not found"
            or ex.args[0] == "error.price-history.io-error"
        ):
            logger.warning("Historical price data not found %s", epic)
            result = pd.DataFrame(
                np.nan,
                columns=pd.MultiIndex.from_tuples(
                    (
                        ("bid", "Open"),
                        ("bid", "High"),
                        ("bid", "Low"),
                        ("bid", "Close"),
                        ("ask", "Open"),
                        ("ask", "High"),
                        ("ask", "Low"),
                        ("ask", "Close"),
                    ),
                ),
                index=pd.DatetimeIndex([], name="DateTime", tz="utc"),
            )
            # raise SkipException after return
        else:
            raise ex
    finally:
        # TODO: need to make this session management RAII
        service.session.close()

    return (
        result["bid"],
        result["ask"],
    )


@symbols.lib_provider(pricelib="{raw_price_lib}")
def create_universe(
    pricelib: Library,
    instruments: pd.DataFrame,
    end_date: pd.Timestamp,
    start_date: pd.Timestamp,
):
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    def get_data(symbol: str):
        return pd.concat(
            (
                pricelib.read(f"{symbol}.bid", date_range=(start_date, end_date)).data,
                pricelib.read(f"{symbol}.ask", date_range=(start_date, end_date)).data,
            ),
            axis=1,
            keys=("bid", "ask"),
        )

    result = pd.concat(
        ((get_data(symbol) for symbol in instruments.index.to_list())),
        axis=1,
        keys=instruments.index.to_list(),
    ).reorder_levels([1, 2, 0], axis=1)
    return (
        result["bid"]["Open"],
        result["bid"]["High"],
        result["bid"]["Low"],
        result["bid"]["Close"],
        result["ask"]["Open"],
        result["ask"]["High"],
        result["ask"]["Low"],
        result["ask"]["Close"],
        ((result["ask"]["Open"] + result["bid"]["Open"]) / 2),
        ((result["ask"]["High"] + result["bid"]["High"]) / 2),
        ((result["ask"]["Low"] + result["bid"]["Low"]) / 2),
        ((result["ask"]["Close"] + result["bid"]["Close"]) / 2),
    )


def get_activity_history(
    from_date,
    to_date,
    svc: IGService | None = None,
):
    """
    get activtiy history and return dataframe per asset
    """

    svc = svc or get_ig_service()

    act = svc.fetch_account_activity_by_date(
        from_date=pd.Timestamp(from_date),
        to_date=pd.Timestamp(to_date),
    )
    svc.session.close()
    act["DateTime"] = (
        pd.to_datetime(act["date"] + " " + act["time"])
        .dt.tz_localize(dateutil.tz.tzlocal())
        .dt.tz_convert("utc")
    )

    return tuple(
        (
            data.set_index("DateTime")
            .sort_index()
            .drop(["date", "time", "period", "epic", "marketName"], axis=1)
            .replace("-", np.nan)
            .astype(
                {
                    "size": float,
                    "stop": float,
                    "limit": float,
                },
                errors="ignore",
            ),
            (epic,),
        )
        for epic, data in iter(act.groupby("epic"))
    )



"""Module for handling environment variables"""

import dataclasses
import json
import os
import pathlib
import typing
from typing import Any, Mapping, Optional

import jinja2
import pandas as pd
import yaml
from jinja2 import Environment, select_autoescape

from tradingo import templates

env = Environment(autoescape=select_autoescape())


class EnvProviderError(Exception):
    """"""


def getbool_(val):
    if isinstance(val, (int, bool)):
        return bool(val)
    if val.lower() in {"true", "yes", "1"}:
        return True
    if val.lower() in {"false", "no", "0"}:
        return False
    raise ValueError(val)


def get_cls(cls):
    if cls is int:
        return int
    if cls is float:
        return float
    if cls is bool:
        return getbool_
    if cls is str:
        return str
    if cls is pd.Timestamp:
        return cls
    if cls is pathlib.Path:
        return cls
    if isinstance(cls, type) and issubclass(cls, EnvProvider):
        return cls
    if isinstance(cls(), typing.Mapping):
        return (
            cls,
            get_cls(typing.get_args(cls)[0]),
            get_cls(typing.get_args(cls)[1]),
        )
    if isinstance(cls(), typing.Iterable):
        return (
            cls,
            get_cls(typing.get_args(cls)[0]),
        )

    raise EnvProviderError(f"Unhandled type '{cls}'")


def _read_dict(cls, val):
    containercls, kcls, vcls = cls
    if isinstance(val, str):
        return containercls(json.loads(val))
    elif isinstance(val, typing.Mapping):
        return val
    else:
        raise EnvProviderError(f"Cannot read mapping '{cls}': '{val}'")


def type_shed(field: dataclasses.Field, val, prefix, env) -> Any:
    cls = get_cls(field.type)

    if isinstance(cls, tuple):
        if len(cls) == 2:
            containercls, vcls = cls
            if isinstance(val, str):
                if isinstance(vcls, tuple):
                    return containercls(
                        _read_dict((vcls[0], *typing.get_args(vcls[0])), v.strip())
                        for v in val.split(";")
                    )
                return containercls(vcls(v) for v in val.split(";"))
            return containercls(val)
        if len(cls) == 3:
            return _read_dict(cls, val)

    if isinstance(cls, type) and issubclass(cls, EnvProvider):
        if isinstance(val, dict):
            return cls(**val, app_prefix=f"{prefix}{field.name}__")
        return cls.from_env(
            app_prefix=f"{prefix}{field.name}_",
            env=env,
        )

    if callable(cls):
        return cls(val)

    raise EnvProviderError(f"Unhandled type {cls} in config")


@dataclasses.dataclass
class EnvProvider:
    app_prefix: Optional[str]

    @classmethod
    def _resolve_args(
        cls,
        env: Mapping[str, str],
        app_prefix: str,
    ):
        out = {}
        for k, v in env.items():
            if not (k.lower().startswith(app_prefix)):
                continue
            k_ = k.lower().replace(app_prefix, "")
            k__ = k_.split("__")[0] if "__" in k_ else k_
            try:
                field = cls.__dataclass_fields__[k__]
            except KeyError:
                raise EnvProviderError(
                    f"Unused config field '{k__}' with value '{v}' for prefix {app_prefix}"
                )
            v_ = type_shed(
                field,
                v,
                app_prefix,
                env,
            )
            out[k__] = v_

        return out

    def to_env(self, env=None):
        env = env if isinstance(env, dict) else os.environ
        for k, v in self.__dict__.items():
            if k == "app_prefix":
                continue
            if isinstance(v, EnvProvider):
                v.to_env(env=env)
                continue
            env[f"{self.app_prefix}{k}".upper()] = str(v)
        return self

    def clean(self, env=None):
        env = env if isinstance(env, dict) else os.environ
        for k, v in self.__dict__.items():
            if k == "app_prefix":
                continue
            if isinstance(v, EnvProvider):
                v.clean(env=env)
                continue
            env.pop(f"{self.app_prefix}{k}".upper(), None)
        return self

    @classmethod
    def from_env(
        cls,
        *,
        app_prefix: str | None = None,
        env: dict[str, Any] | None = None,
        override_default_env: bool = True,
    ):

        try:
            app_prefix = app_prefix or getattr(cls, "app_prefix")
        except AttributeError as ex:
            raise EnvProviderError(
                "app_prefix must be passed or set on the config object"
            ) from ex

        app_prefix = (app_prefix + "_").lower()

        env = env or dict(os.environ)

        resolved_args = cls._resolve_args(env, app_prefix)

        default_env = cls._resolve_args(os.environ, app_prefix)

        if override_default_env:
            resolved_args.update(
                (k, v) for k, v in default_env.items() if k in resolved_args
            )

        try:
            return cls(**resolved_args, app_prefix=app_prefix)
        except TypeError as ex:
            if "missing" in ex.args[0]:
                raise EnvProviderError(
                    f"Missing settings:{ex.args[0].split(':')[-1]} for {app_prefix}"
                ) from ex

            if "unexpected" in ex.args[0]:
                raise EnvProviderError(
                    f"Unknown settings:{ex.args[0].split('argument')[-1]}"
                ) from ex
            raise ex

    @classmethod
    def from_file(cls, variables: str | pathlib.Path):
        variables = pathlib.Path(variables)
        rendered = jinja2.Template(variables.read_text()).render(**os.environ)
        if variables.suffix == ".json":
            return cls.from_env(env=json.loads(rendered))
        if variables.suffix == ".yaml":
            return cls.from_env(env=yaml.safe_load(rendered))
        raise ValueError(variables.suffix)


@dataclasses.dataclass
class IGTradingConfig(EnvProvider):
    """IG Trading configuration"""

    password: str
    username: str
    api_key: str
    acc_type: str
    app_prefix = "IG_SERVICE"


@dataclasses.dataclass
class TradingoConfig(EnvProvider):
    """Tradingo configuration"""

    config_home: pathlib.Path
    arctic_uri: str
    templates: pathlib.Path = pathlib.Path(templates.__file__).parent
    include_instruments: bool = False
    app_prefix = "TP"


@dataclasses.dataclass
class SMTPConfig(EnvProvider):
    """SMTP server configuration"""

    server_uri: str
    port: int
    username: str
    password: str
    app_prefix = "SMTP"



"""Tradingo CLI"""

from __future__ import annotations

import argparse
import logging
import os
import pathlib
import re

import pandas as pd

from tradingo.api import Tradingo
from tradingo.config import read_config_template
from tradingo.dag import DAG
from tradingo.settings import IGTradingConfig, TradingoConfig

logger = logging.getLogger(__name__)


def int_or_bool(val: str) -> int | bool:
    if val.lower() in {"true", "yes"}:
        return True
    if val.lower() == {"false", "no"}:
        return False
    if not val.isnumeric():
        raise ValueError("Invalid value for int or bool: {val}")
    return int(val)


def cli_app() -> argparse.ArgumentParser:
    """Tradingo CLI app."""

    app = argparse.ArgumentParser("tradingo-tasks")

    app.add_argument(
        "--auth",
        type=lambda i: IGTradingConfig.from_env(
            env=read_config_template(pathlib.Path(i), os.environ),
            override_default_env=False,
        ).to_env(),
        required=False,
    )
    app.add_argument(
        "--config",
        type=lambda i: read_config_template(pathlib.Path(i), os.environ),
        required=True,
    )

    entity = app.add_subparsers(dest="entity", required=True)
    universe = entity.add_parser("universe")
    universe_subparsers = universe.add_subparsers(dest="universe_action", required=True)
    _ = universe_subparsers.add_parser("list")
    uni_show = universe_subparsers.add_parser("show")

    uni_show.add_argument("name")

    task = entity.add_parser("task")

    task_subparsers = task.add_subparsers(dest="list_action", required=True)
    run_tasks = task_subparsers.add_parser("run")
    run_tasks.add_argument("task")
    run_tasks.add_argument("--with-deps", default=False, type=int_or_bool)
    run_tasks.add_argument("--start-date", type=pd.Timestamp, required=False)
    run_tasks.add_argument("--end-date", type=pd.Timestamp, required=False)
    run_tasks.add_argument("--force-rerun", action="store_true", default=True)
    run_tasks.add_argument("--dry-run", action="store_true")
    run_tasks.add_argument("--clean", action="store_true")
    run_tasks.add_argument("--skip-deps", type=re.compile)

    _ = task_subparsers.add_parser("list")

    return app


def handle_tasks(args, arctic):
    """inspect or run Tradingo tasks."""

    if args.list_action == "list":
        graph = DAG.from_config(
            args.config,
        )

        graph.print()

    elif args.list_action == "run":
        graph = DAG.from_config(
            args.config,
        )

        graph.update_state()

        try:
            extra_kwargs = {}
            if args.start_date:
                extra_kwargs["start_date"] = args.start_date
            if args.end_date:
                extra_kwargs["end_date"] = args.end_date
            if args.clean:
                extra_kwargs["clean"] = args.clean
            out = graph.run(
                args.task,
                run_dependencies=args.with_deps,
                force_rerun=args.force_rerun,
                arctic=arctic,
                dry_run=args.dry_run,
                skip_deps=args.skip_deps,
                **extra_kwargs,
            )
            if args.dry_run:
                print(out)
        finally:
            graph.serialise_state()

    else:
        raise ValueError(args.list_action)


def handle_universes(args, api: Tradingo):
    """inspect Tradingo's universe by inspecting the instruments in DB."""

    if args.universe_action == "list":
        for item in api.instruments.list():
            print(item)

    elif args.universe_action == "show":
        print(api.instruments[args.name]())

    elif args.universe_action == "prices":
        print(api.instruments[args.name]())

    else:
        raise ValueError(args.universe_action)


def main():
    """Tradingo CLI entrypoint"""

    envconfig = TradingoConfig.from_env().to_env()
    args = cli_app().parse_args()
    IGTradingConfig.from_env().to_env()
    envconfig.to_env()

    arctic = Tradingo(envconfig.arctic_uri)
    if args.entity == "task":
        handle_tasks(args, arctic)

    elif args.entity == "universe":
        handle_universes(args, api=arctic)

    else:
        raise ValueError(args.entity)


if __name__ == "__main__":
    logging.getLogger("tradingo").setLevel(logging.INFO)
    main()



"""static instruments data accessors."""

import logging
from typing import Literal, Optional

import pandas as pd
from yfinance import Ticker

from tradingo.sampling.ig import get_ig_service
from tradingo.sampling.yf import currency_to_symbol

logger = logging.getLogger(__name__)


Provider = Literal[
    "alpha_vantage",
    "cboe",
    "fmp",
    "intrinio",
    "polygon",
    "tiingo",
    "tmx",
    "tradier",
    "yfinance",
]


def download_instruments(
    *,
    index_col: Optional[str] = None,
    html: Optional[str] = None,
    file: Optional[str] = None,
    tickers: Optional[list[str]] = None,
    epics: Optional[list[str]] = None,
):
    if file:
        return (
            pd.read_csv(
                file,
                index_col=index_col,
            ).rename_axis("Symbol"),
        )

    if html:
        return (pd.read_html(html)[0].set_index(index_col).rename_axis("Symbol"),)

    if tickers:
        if isinstance(tickers, str) and tickers.startswith("http"):
            tickers = pd.read_html(tickers)[0][index_col]
        if isinstance(tickers, str) and tickers.startswith("file"):
            breakpoint()
            tickers = pd.read_csv(tickers.split("://")[-1])[index_col]
        return (
            (
                pd.DataFrame(
                    {t: Ticker(currency_to_symbol(t)).get_info() for t in tickers}
                )
                .transpose()
                .rename_axis("Symbol")
            ),
        )

    if epics:
        service = get_ig_service()

        return (
            pd.DataFrame((service.fetch_market_by_epic(e)["instrument"] for e in epics))
            .set_index("epic")
            .rename_axis("Symbol", axis=0),
        )

    raise ValueError(file)



from unittest.mock import patch

import pandas as pd
import pytest

from tradingo.sampling.yf import (
    _align_series,
    _get_ticker,
    adjust_fx_series,
    currency_to_symbol,
    sample_equity,
    symbol_to_currency,
)


def test_handle_currency_ticker_valid_ticker():
    symbols = "USDJPY=X"
    result = symbol_to_currency(symbols)
    assert (
        result == "USDJPY"
    ), "The ticker should be correctly converted to currency pair."


def test_handle_currency():
    ticker = "GBPUSD"
    result = currency_to_symbol(ticker)
    assert result == "GBPUSD=X", "The ticker is a currency pair, whould end by =X."


def test_handle_not_a_currency():
    ticker = "NOTCCY"
    result = currency_to_symbol(ticker)
    assert result == "NOTCCY", "The ticker is not a currency pair."


def test_get_ticker_logs():
    t = _get_ticker("USDEUR")
    assert t.endswith("=X")


@patch("tradingo.sampling.yf.yf.download")
def test_sample_equity_calls_download(mock_download):
    mock_download.return_value = pd.DataFrame({"Close": [1, 2]})
    (df,) = sample_equity("USDEUR", "2020-01-01", "2020-01-02")
    assert isinstance(df, pd.DataFrame)
    mock_download.assert_called_once()


def test_adjust_fx_series_basic():
    df = pd.DataFrame({"EURUSD": [2.0, 4.0], "USDEUR": [0.5, 0.25]})
    out = adjust_fx_series(df, ref_ccy="USD", add_self=True)
    expected = pd.DataFrame({"EUR": [2.0, 4.0], "USD": [1.0, 1.0]})
    pd.testing.assert_frame_equal(out, expected)


def test_align_series_basic():
    s = pd.Series([1, 2, 3])
    o = 2
    s1, s2 = _align_series(s, o)
    assert (s2 == 2).all()
    s1, s2 = _align_series(s)
    assert (s2 == 1).all()


if __name__ == "__main__":
    pytest.main([__file__])



import os
import textwrap

import pytest

from tradingo import config, dag
from tradingo.settings import TradingoConfig


@pytest.fixture
def config_home(tmp_path):
    (tmp_path / "configs").mkdir(exist_ok=True)
    (tmp_path / "configs" / "signals").mkdir(exist_ok=True)
    (tmp_path / "configs" / "universes").mkdir(exist_ok=True)

    (tmp_path / "configs" / "myconfig.yaml").write_text(
        textwrap.dedent(
            """\
        {% set universe_name = 'myuniverse' %}
        prices:
            include: "file://{{ TP_TEMPLATES }}/instruments/ig-trading.yaml"
            variables:
                include: "file://{{ TP_CONFIG_HOME }}/universes/{{ universe_name }}.yaml"
                start_date: "{{ data_interval_start }}"
                end_date: "{{ data_interval_end }}"
        signals:
            include: "file://{{ TP_CONFIG_HOME }}/signals/mysignals.yaml"
            variables:
                include: "file://{{ TP_CONFIG_HOME }}/universes/{{ universe_name }}.yaml"
                universe_name: "{{ universe_name }}"
                bid_close: "prices/{{ universe_name }}.bid.close"
                ask_close: "prices/{{ universe_name }}.ask.close"
                frequency: "15min"
        portfolio:
            include: "file://{{ TP_TEMPLATES }}/portfolio_construction.yaml"
            variables:
                close: "prices/{{ universe_name }}.mid.close"
                ask_close: "prices/{{ universe_name }}.ask.close"
                bid_close: "prices/{{ universe_name }}.bid.close"
                backtest:
                price_ffill_limit: 5
                universe_name: "{{ universe_name }}"
                name: intraday
                model_weights:
                    "intraday_momentum.{{ universe_name }}": 1.0
                portfolio:
                aum: 10000
                multiplier: 0.5
        """
        )
    )
    (tmp_path / "configs" / "signals" / "mysignals.yaml").write_text(
        textwrap.dedent(
            """\
            "signals.intraday_momentum.myuniverse": 
                function: func
                symbols_in: []
                symbols_out: []
                depends_on: ["sample.{{ universe_name }}"]
                params: {}
            """
        )
    )
    (tmp_path / "configs" / "universes" / "myuniverse.yaml").write_text(
        textwrap.dedent(
            """\
            universe_name: im-multi-asset-3
            raw_prices_lib: prices_igtrading
            epics:
                - "CC.D.NG.UMP.IP"
                - "IX.D.SPTRD.IFS.IP"
                - "CC.D.LCO.UMP.IP"
                - "IR.D.10YEAR100.FWM2.IP"
                - "CC.D.CC.UMP.IP"
                - "CC.D.S.UMP.IP"
                - "IX.D.NASDAQ.IFS.IP"
                - "CS.D.CFPGOLD.CFP.IP"
            interval: 15min
            start_date: "2017-01-01 00:00:00+00:00"
            end_date: "{{ data_interval_end }}"
            """
        )
    )

    return tmp_path / "configs"


def test_config(config_home):
    env = TradingoConfig.from_env(
        env={
            "TP_CONFIG_HOME": str(config_home),
            "TP_ARCTIC_URI": "mem://",
        }
    )
    env.to_env()

    out = config.read_config_template(
        config_home / "myconfig.yaml",
        variables=os.environ,
    )

    dag.DAG.from_config(out)



from test.tradingo.utils import close_position

import numpy as np
import pandas as pd
import pytest

from tradingo import backtest


def test_backtest_integration(tradingo):
    pd.concat(
        backtest.backtest(
            portfolio=tradingo.portfolio.model.raw.shares(),
            bid_close=tradingo.prices.close(),
            ask_close=tradingo.prices.close(),
            dividends=None,
        ),
        keys=("portfolio", *(f"instrument.{i}" for i in backtest.BACKTEST_FIELDS)),
        axis=1,
    )


@pytest.mark.parametrize(
    "prices_,portfolio_,unrealised_pnl,realised_pnl",
    [
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            pd.DataFrame(
                1,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            "(prices_*portfolio_).squeeze().diff()",
            "pd.Series(0.0, index=prices_.index)",
        ),
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            pd.DataFrame(
                range(0, 260),
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ),
            "pd.Series([np.nan, np.nan, *(prices_.diff()*portfolio_.shift()).squeeze().iloc[2:]], index=prices_.index)",
            "pd.Series(0.0, index=prices_.index)",
        ),
        (
            pd.DataFrame(
                1 + 0.01,
                index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                columns=["ABCD"],
            ).cumprod(),
            close_position(
                pd.DataFrame(
                    1,
                    index=pd.bdate_range("2020-01-01", periods=260, freq="B"),
                    columns=["ABCD"],
                ).cumprod(),
            ),
            "pd.Series([*(prices_.diff()*portfolio_.shift()).squeeze().iloc[0:-1], np.nan], index=prices_.index)",
            "pd.Series([*(0 for _ in range(0, 259)), 12.280985478055433], index=prices_.index)",
        ),
    ],
)
def test_backtest_smoke(tradingo, prices_, portfolio_, unrealised_pnl, realised_pnl):
    dividends = pd.DataFrame(0, columns=prices_.columns, index=prices_.index)
    bt = pd.concat(
        backtest.backtest(
            portfolio=portfolio_,
            bid_close=prices_,
            ask_close=prices_,
            dividends=dividends,
        ),
        keys=("portfolio", *(f"instrument.{i}" for i in backtest.BACKTEST_FIELDS)),
        axis=1,
    )

    actual_unrealised = bt["instrument.unrealised_pnl"].squeeze().diff()
    expected_unrealised = (
        eval(unrealised_pnl) if isinstance(unrealised_pnl, str) else unrealised_pnl
    )

    pd.testing.assert_series_equal(
        actual_unrealised,
        expected_unrealised.astype("float32"),
        check_names=False,
        check_freq=False,
        rtol=1e-4,
    )
    actual_realised = bt["instrument.realised_pnl"].squeeze().diff().fillna(0.0)
    expected_realised = (
        eval(realised_pnl) if isinstance(realised_pnl, str) else realised_pnl
    )
    pd.testing.assert_series_equal(
        actual_realised,
        expected_realised.astype("float32"),
        check_names=False,
        check_freq=False,
        rtol=1e-4,
    )


@pytest.mark.parametrize(
    "position, prices, unrealised_pnl, realised_pnl, total_pnl",
    (
        (
            pd.DataFrame(
                [0, 1, 1, 1, 0, 0],
                columns=["ABC"],
                index=pd.bdate_range("2024-11-24", periods=6),
            ),
            pd.DataFrame(
                [10, 10, 11, 12, 13, 14],
                columns=["ABC"],
                index=pd.bdate_range("2024-11-24", periods=6),
            ),
            pd.DataFrame(
                [np.nan, 0, 1, 2, np.nan, np.nan],
                columns=["ABC"],
                index=pd.bdate_range("2024-11-24", periods=6),
            ),
            pd.DataFrame(
                [0, 0, 0, 0, 3, 3],
                columns=["ABC"],
                index=pd.bdate_range("2024-11-24", periods=6),
            ),
            pd.DataFrame(
                [0, 0, 1, 2, 3, 3],
                columns=["ABC"],
                index=pd.bdate_range("2024-11-24", periods=6),
            ),
        ),
    ),
)
def test_backtest_all_closed(
    position,
    prices,
    unrealised_pnl,
    realised_pnl,
    total_pnl,
):
    bt = pd.concat(
        backtest.backtest(
            portfolio=position,
            bid_close=prices,
            ask_close=prices,
            dividends=None,
        ),
        keys=("portfolio", *(f"instrument.{i}" for i in backtest.BACKTEST_FIELDS)),
        axis=1,
    )

    pd.testing.assert_frame_equal(
        bt["instrument.unrealised_pnl"],
        unrealised_pnl,
        check_dtype=False,
    )
    pd.testing.assert_frame_equal(
        bt["instrument.realised_pnl"],
        realised_pnl,
        check_dtype=False,
    )
    pd.testing.assert_frame_equal(
        bt["instrument.total_pnl"],
        total_pnl,
        check_dtype=False,
    )



import logging
import re
from typing import Optional

import numpy as np
import pandas as pd
from arcticdb.arctic import Library

from tradingo import symbols

logger = logging.getLogger(__name__)


@symbols.lib_provider(signals="signals")
def portfolio_construction(
    signals: Library,
    close: pd.DataFrame,
    model_weights: dict[str, float],
    multiplier: float,
    aum: float,
    start_date: pd.Timestamp,
    end_date: pd.Timestamp,
    instruments: Optional[pd.DataFrame] = None,
    default_instrument_weight: float = 1.0,
    instrument_weights: Optional[dict] = None,
):
    """Catch all portfolio construction function for basic
    portfolio construction routines

    :param signals: library of where to find signals
    :param close: dataframe of close prices
    :param model_weights: dictionary of model weights. Keys correspond
        to a symbol in the library
    :param multiplier: some scalar to multiply weights by
    :param aum: notional aum value
    :param start_date: the start date to use
    :param end_date: the end date to use
    :param instruments: optional dataframe of instruments for asset type weights
    :param instrument_weights: loading to apply to instruments
    """
    instrument_weights = instrument_weights or {}

    weights = pd.Series(multiplier, index=close.columns)

    if instruments is not None:

        instruments["Symbol"] = instruments.index

        for key, weights_config in instrument_weights.items():
            if key not in instruments.columns:
                continue

            weights = weights * instruments.apply(
                lambda i: weights_config.get(i[key], default_instrument_weight),
                axis=1,
            )
            logger.info("Weights: %s", weights)

    signal_value = (
        pd.concat(
            (
                weight
                * signals.read(
                    model_name,
                    date_range=(start_date, end_date),
                ).data
                for model_name, weight in model_weights.items()
            ),
            keys=model_weights,
            axis=1,
        )
        .transpose()
        .groupby(level=[1])
        .sum()
        .transpose()
        .ffill()
        .fillna(0)
    )

    positions = signal_value.reindex_like(close).ffill()

    pct_position = positions.div(positions.transpose().sum(), axis=0).fillna(0.0)
    for col in pct_position.columns:
        pct_position.loc[
            pct_position.index == pct_position.first_valid_index(), col
        ] = 0.0

    share_position = (pct_position * aum) / close

    return (
        pct_position,
        share_position,
        positions,
        pct_position.round(decimals=2),
        share_position.round(),
        positions.round(),
        signal_value,
    )


def portfolio_optimization(
    close: pd.DataFrame,
    factor_returns: pd.DataFrame,
    optimizer_config: dict,
    rebalance_rule: str,
    min_periods: int,
    aum: float,
):
    import riskfolio as rf

    def get_weights(
        returns,
        factors,
    ):
        port = rf.Portfolio(returns=returns)

        port.assets_stats(method_mu="hist", method_cov="ledoit")
        port.lowerret = 0.00056488 * 1.5

        port.factors = factors

        port.factors_stats(
            method_mu="hist",
            method_cov="ledoit",
            feature_selection="PCR",
        )

        w = port.optimization(
            model="FM",
            rm="MV",
            obj="Sharpe",
            hist=False,
        )
        return (
            w.squeeze() if w is not None else pd.Series(np.nan, index=returns.columns)
        )

    asset_returns = close.pct_change().dropna()
    factor_returns = close.pct_change().dropna().reindex(asset_returns.index)

    data = []
    for i, _ in enumerate(asset_returns.index):
        if i < min_periods:
            data.append(
                pd.Series(np.nan, index=asset_returns.columns).to_frame().transpose()
            )
            continue

        ret_subset = asset_returns.iloc[:i]
        data.append(
            get_weights(ret_subset, factor_returns.loc[ret_subset.index])
            .to_frame()
            .transpose()
        )

    pct_position = pd.concat(data, keys=asset_returns.index).droplevel(1)
    share_position = (pct_position * aum) / close

    return (pct_position, share_position)


def instrument_ivol(close, provider, **kwargs):
    pct_returns = np.log(close / close.shift())

    ivols = []

    for symbol in pct_returns.columns:
        universe = pct_returns.drop(symbol, axis=1)

        def vol(uni):
            return (1 - (1 + uni).prod(axis=1).pow(1 / 100)).ewm(10).std()

        ivol = vol(pd.concat((universe, pct_returns[symbol]), axis=1)) - vol(universe)
        ivols.append(ivol.rename(symbol))

    return (pd.concat(ivols, axis=1).rename_axis("Symbol"),)


def _parse_ticker(t: str):
    match = re.match(r".*\(([A-Z]{4})\)", t)
    if match is not None:
        return match.groups()[0]
    return t


def position_from_trades(
    close: pd.DataFrame,
    aum: float,
    trade_file: str,
):
    trades = (
        pd.read_csv(trade_file, parse_dates=["Date"])
        .dropna(axis=0, how="all")
        .sort_values(["Date"])
    )
    trades = trades[
        trades["Order type"].isin(["AtBest", "Quote and Deal"])
        & trades["Order status"].eq("Completed")
    ]
    trades["Ticker"] = trades["Investment"].apply(_parse_ticker) + ".L"
    position_shares = (
        trades.set_index(["Date", "Ticker"])
        .groupby(["Date", "Ticker"])
        .sum()
        .unstack()["My units"]
        .fillna(0.0)
        .cumsum()
        .reindex_like(
            close,
        )
        .ffill()
        .fillna(0.0)
    )

    position_pct = (position_shares * close) / aum
    return (
        position_pct,
        position_shares,
    )


def point_in_time_position(positions: pd.DataFrame):
    return ((positions).iloc[-1:,],)



from __future__ import annotations

import contextlib
import inspect
import re
from typing import Any, Optional

import arcticdb as adb
import pandas as pd

import tradingo.utils

READ_SIG = inspect.signature(adb.arctic.Library.read)


class _Read:
    def __init__(
        self,
        path_so_far,
        library: adb.library.Library,
        assets,
        common_args,
        common_kwargs,
        root: Tradingo,
    ):
        self._path_so_far = path_so_far
        self._library = library
        self._assets = assets
        self._path = ".".join(self._path_so_far)
        self._common_args = common_args
        self._common_kwargs = common_kwargs
        self._root = root

    def __dir__(self):
        return [*self.list(), *super().__dir__()]

    def __repr__(self):
        return f'Namespace("{self._path}")'

    def __getattr__(self, symbol):
        return self.__class__(
            (*self._path_so_far, symbol),
            library=self._library,
            assets=self._assets,
            common_args=self._common_args,
            common_kwargs=self._common_kwargs,
            root=self._root,
        )

    def __getitem__(self, symbol):
        return self.__getattr__(symbol)

    def __call__(self, *args, **kwargs) -> pd.DataFrame:
        operations = ["merge", "transpose", "concat", "with_instrument_details"]
        operation, index = next(
            (
                (elem, i)
                for i, elem in enumerate(self._path_so_far)
                if elem in operations
            ),
            (None, len(self._path_so_far)),
        )
        part_one_path = self._path_so_far[0:index]
        part_two_path = self._path_so_far[index + 1 :]

        lib_kwargs = {
            k: v for k, v in kwargs.items() if k in READ_SIG.parameters.keys()
        }

        callback_kwargs = {}

        def get_callback(operation):
            for i in (pd.DataFrame, tradingo.utils, pd):
                try:
                    return getattr(i, operation)
                except AttributeError:
                    continue
            raise AttributeError(operation)

        if operation:
            callback = get_callback(operation)
            callback_sig = inspect.signature(callback)
            callback_kwargs = {
                k: v for k, v in kwargs.items() if k in callback_sig.parameters.keys()
            }

        def get_data_at_path(path, kw) -> pd.DataFrame:
            kw.setdefault(
                "columns",
                (
                    self._assets
                    if all(i in path for i in ("backtest", "portfolio"))
                    else None
                ),
            )

            for k in self._common_kwargs:
                kw.pop(k, None)
            return self._library.read(
                ".".join(path),
                *args,
                *self._common_args,
                **kw,
                **self._common_kwargs,
            ).data

        lhs = get_data_at_path(part_one_path, lib_kwargs)

        if operation:
            callback_args: (
                tuple[pd.DataFrame]
                | tuple[pd.DataFrame, pd.DataFrame]
                | tuple[tuple[pd.DataFrame, pd.DataFrame]]
            ) = (lhs,)
            if part_two_path:
                callback_lib = part_two_path[0]
                callback_args = (
                    lhs,
                    self.__class__(
                        (
                            *self._root._get_path_so_far(callback_lib),
                            *part_two_path[1:],
                        ),
                        getattr(self._root, (callback_lib))._library,
                        self._assets,
                        self._common_args,
                        self._common_kwargs,
                        self._root,
                    )(*args, **kwargs),
                )

            if operation == "concat":
                callback_args = (callback_args,)  # type: ignore

            callback = get_callback(operation)
            callback_sig = inspect.signature(pd.DataFrame)
            return callback(*callback_args, **callback_kwargs)
        return lhs

    def update(self, *args, **kwargs):
        self._library.update(self._path, *args, **kwargs)

    def list(self, *args, **kwargs):
        if self._path_so_far:
            regex = re.escape(".".join((self._path_so_far)) + ".")
            kwargs["regex"] = regex + kwargs.setdefault("regex", "")
        return list(
            dict.fromkeys(
                [
                    i.replace(
                        f"{'.'.join(self._path_so_far)}." if self._path_so_far else "",
                        "",
                    ).split(".")[0]
                    for i in self._library.list_symbols(*args, **kwargs)
                ]
            )
        )

    def head(self, *args, **kwargs):
        return self._library.head(self._path, *args, **kwargs).data

    def tail(self, *args, **kwargs):
        return self._library.tail(self._path, *args, **kwargs).data

    def exists(self):
        """Return true if symbol exists"""
        return bool(self._library.list_symbols(regex=f"^{re.escape(self._path)}$"))


class Tradingo(adb.Arctic):
    def __init__(
        self,
        *args,
        provider: Optional[str] = None,
        universe: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)
        self._provider = provider
        self._universe = universe
        self._context_args = ()
        self._context_kwargs: dict[str, Any] = {}

    @contextlib.contextmanager
    def common_args(self, *args, **kwargs):
        try:
            self._context_args = args
            self._context_kwargs = kwargs
            yield self
        finally:
            self._context_args = ()
            self._context_kwargs = {}

    def _get_path_so_far(self, library):
        path_so_far = []
        if library == "instruments":
            return path_so_far
        if self._provider:
            path_so_far.append(self._provider)
        if self._universe:
            path_so_far.append(self._universe)
        return path_so_far

    def __getattr__(self, library):
        if library in self.list_libraries():
            path_so_far = self._get_path_so_far(library)
            if library == "instruments":
                return _Read(
                    library=self.get_library(library),
                    path_so_far=path_so_far,
                    assets=[],
                    common_args=(),
                    common_kwargs={},
                    root=self,
                )

            assets = []
            if self._universe:
                assets = getattr(self.instruments, self._universe)().index.to_list()

            return _Read(
                library=self.get_library(library),
                path_so_far=path_so_far,
                assets=assets,
                common_args=self._context_args,
                common_kwargs=self._context_kwargs,
                root=self,
            )

        return super().__getattr__(library)

    def __dir__(self):
        return [*self.list_libraries(), *super().__dir__()]


class VolSurface(Tradingo):
    def get(
        self,
        symbol: str,
        start_date: pd.Timestamp = pd.Timestamp("1970-01-01 00:00+00:00"),
        end_date: pd.Timestamp = pd.Timestamp.now("utc"),
        **kwargs,
    ):
        with self.common_args(date_range=(start_date, end_date)):
            futures = (
                pd.concat(
                    (self.futures.cboe.VX.expiration(), self.futures.cboe.VX.price()),
                    axis=1,
                    keys=("expiration", "price"),
                )
                .stack()
                .astype({"expiration": "datetime64[ns]"})
                .reset_index()
                .set_index(["timestamp", "symbol"])
            )

            library = getattr(self.options.cboe, symbol)
            option_chains = pd.concat(
                (
                    library.expiration(),
                    library.option_type(),
                    library.strike(),
                    library.bid(),
                    library.ask(),
                    library.implied_volatility(),
                    library.delta(),
                    library.vega(),
                    library.gamma(),
                    library.theta(),
                    library.rho(),
                ),
                axis=1,
                keys=(
                    "expiration",
                    "option_type",
                    "strike",
                    "bid",
                    "ask",
                    "implied_volatility",
                    "delta",
                    "vega",
                    "gamma",
                    "theta",
                    "rho",
                ),
            ).stack(future_stack=True)

        return option_chains.merge(
            futures, on=["timestamp", "expiration"], how="left"
        ).set_index(["expiration", "option_type", "strike"], append=True)



"""Tradingo config."""

import json
import pathlib
from typing import Any, MutableMapping, Optional

import jinja2
import pandas as pd
import yaml


class ConfigLoadError(Exception):
    """raised when loading config incorrectly."""


def read_config_template(
    filepath: pathlib.Path,
    variables: dict[str, Any] | MutableMapping,
    path_so_far: Optional[list[str]] = None,
) -> dict[str, Any]:
    """read a config template and populate it with values"""
    filepath = pathlib.Path(filepath)

    path_so_far = path_so_far or []

    try:
        renderedtext = jinja2.Template(filepath.read_text(encoding="utf-8")).render(
            **variables
        )
    except Exception as ex:
        raise ex

    try:
        if filepath.suffix == ".json":
            return process_includes(
                json.loads(renderedtext),
                variables,
                path_so_far=path_so_far,
            )

        if filepath.suffix == ".yaml":
            return process_includes(
                yaml.safe_load(renderedtext),
                variables,
                path_so_far=path_so_far,
            )

    except jinja2.TemplateSyntaxError as ex:
        raise ConfigLoadError(f"Error reading config template: {filepath}") from ex
    except FileNotFoundError as ex:
        raise ConfigLoadError(
            f"Error rendering include: {filepath=} {path_so_far=}"
        ) from ex

    raise ValueError(f"Unhandled file type: '{filepath.suffix}'")


def process_includes(
    config: dict[str, Any],
    variables,
    path_so_far: Optional[list[str]] = None,
) -> dict[str, Any]:
    """bake nested configs via 'include' pattern."""

    out = {}

    path_so_far = path_so_far or []
    for key, value in config.items():
        path_so_far.append(key)
        if isinstance(value, dict) and "include" in value:
            value = process_includes(value, variables, path_so_far)

            protocol, path = value["include"].split("://")

            if protocol == "file" and (
                path.endswith(".json") or path.endswith(".yaml")
            ):
                try:
                    incvalue = read_config_template(
                        pathlib.Path(path),
                        variables={
                            **value.get("variables", {}),
                            **variables,
                        },
                        path_so_far=path_so_far,
                    )
                    value = value.copy()
                    value.update(incvalue)
                    value.pop("include", None)
                    value.pop("variables", None)
                except ConfigLoadError as ex:
                    raise ConfigLoadError(
                        f"Error processing include at '{value}' {path_so_far=}"
                    ) from ex
            elif protocol == "file" and (
                path.endswith(".csv") or path.endswith(".parquet")
            ):
                df = pd.read_csv(path)
                if query := value.get("query"):
                    df = df.query(query)
                incvalue = df[value["column"]].to_list()
                value = incvalue

            elif protocol.startswith("http"):
                df = pd.read_html(value["include"])[0]
                if query := value.get("query"):
                    df = df.query(query)
                incvalue = df[value["column"]].to_list()
                value = incvalue

            else:
                raise ValueError(
                    f"Unsupported protocol: '{protocol}' at '{key}' for '{path}'"
                )

        elif isinstance(value, dict):
            value = process_includes(value, variables, path_so_far)

        out[key] = value

    return out



import dataclasses
import json
import os
import pathlib
from unittest import mock

import pytest

from tradingo.settings import EnvProvider, EnvProviderError


def test_env_provider(tmp_path):
    @dataclasses.dataclass
    class Settings1(EnvProvider):
        parameter: int

    env = {"TRADINGO_PARAMETER": 1}
    prefix = "TRADINGO"

    settings = Settings1.from_env(app_prefix=prefix, env=env)
    assert settings.parameter == 1

    env = {"TRADINGO_PARAMETER": "1"}
    with mock.patch.dict(os.environ, env):
        s = Settings1.from_env(app_prefix=prefix, env=env)
    assert s.parameter == 1

    with pytest.raises(EnvProviderError):
        s = Settings1.from_env(env=env)

    @dataclasses.dataclass
    class Settings2(EnvProvider):
        parameter: bool
        app_prefix = "TRADINGO"

    env = {"TRADINGO_PARAMETER": "true"}
    s = Settings2.from_env(env=env)
    assert s.parameter

    env = {"TRADINGO_PARAMETER": "false"}
    s = Settings2.from_env(env=env)
    assert not s.parameter

    @dataclasses.dataclass
    class Settings3(EnvProvider):
        param1: int
        param2: float
        param3: bool
        param4: str
        app_prefix = "APP"

    env = {"TRADINGO_PARAMETER": "1"}

    with pytest.raises(
        EnvProviderError,
        match="Missing settings: 'param1', 'param2', 'param3', and 'param4'",
    ):
        s = Settings3.from_env(env=env)

    s3 = Settings3.from_env(
        env={
            "app_param1": 1,
            "app_param2": 1.5,
            "app_param3": False,
            "app_param4": "value",
        }
    )

    assert s3.param1 == 1
    assert s3.param2 == 1.5
    assert not s3.param3
    assert s3.param4 == "value"

    with pytest.raises(
        EnvProviderError,
        match="Unused config field 'unknown' with value 'value' for prefix app_",
    ):
        s3 = Settings3.from_env(
            env={
                "app_param1": 1,
                "app_param2": 1.5,
                "app_param3": False,
                "app_param4": "value",
                "app_unknown": "value",
            }
        )

    @dataclasses.dataclass
    class Settings4(EnvProvider):
        param1: Settings3
        param2: str
        app_prefix = "APP"

    env_flat = {
        "app_param1__param1": 1,
        "app_param1__param2": 1.5,
        "app_param1__param3": False,
        "app_param1__param4": "value",
        "app_param2": "value",
    }

    s4 = Settings4.from_env(env=env_flat)

    assert s4.param1 == Settings3(
        param1=1,
        param2=1.5,
        param3=False,
        param4="value",
        app_prefix="app_param1__",
    )

    env = {
        "app_param1": {
            "param1": 1,
            "param2": 1.5,
            "param3": False,
            "param4": "value",
        },
        "app_param2": "value",
    }

    s4 = Settings4.from_env(env=env)

    assert s4.param1 == Settings3(
        param1=1,
        param2=1.5,
        param3=False,
        param4="value",
        app_prefix="app_param1__",
    )

    vars = pathlib.Path(tmp_path) / "variables.json"
    vars.write_text(json.dumps(env_flat))
    s4_file = Settings4.from_file(vars)

    assert s4_file == s4

    vars.write_text(json.dumps(env))
    s4_file = Settings4.from_file(vars)
    assert s4_file == s4

    @dataclasses.dataclass
    class Settings5(EnvProvider):
        param1: list[str]
        param2: tuple[str]
        param3: list[dict[str, str]]
        param4: dict[str, str]
        app_prefix = "APP"

    env = {
        "app_param1": ["val", "val2"],
        "app_param2": ("val",),
        "app_param3": [{"key": "val"}],
        "app_param4": {"key": "val"},
    }

    vars.write_text(json.dumps(env))
    s5_file = Settings5.from_file(vars)

    env = {
        "app_param1": "val;val2",
        "app_param2": "val",
        "app_param3": '{"key": "val"}',
        "app_param4": '{"key": "val"}',
    }
    s5_env = Settings5.from_env(env=env)
    assert s5_file == s5_env
    assert s5_env == Settings5(
        param1=["val", "val2"],
        param2=("val",),
        param3=[{"key": "val"}],
        param4={"key": "val"},
        app_prefix="app_",
    )

    s5 = Settings5.from_env(env=env)
    assert s5 == s5_env

    env = {}
    s3.to_env(env)
    assert env == {
        "APP_PARAM1": "1",
        "APP_PARAM2": "1.5",
        "APP_PARAM3": "False",
        "APP_PARAM4": "value",
    }

    env = {}
    s4.to_env(env)
    assert env == {
        "APP_PARAM1__PARAM1": "1",
        "APP_PARAM1__PARAM2": "1.5",
        "APP_PARAM1__PARAM3": "False",
        "APP_PARAM1__PARAM4": "value",
        "APP_PARAM2": "value",
    }



"""Yahoo Finance data provider."""

import logging

import pandas as pd
import pycountry
import yfinance as yf
from arcticdb.version_store.library import Library

from tradingo import symbols

logger = logging.getLogger(__name__)


_CCY_CODES = {c.alpha_3 for c in pycountry.currencies}


class ProviderDataError(Exception):
    """Raised if there is an issue with data availability
    or gathering from provider."""


def currency_to_symbol(maybe_currency: str) -> str:
    """Convert a currency pair to a YF currency symbol, prepending '=X'."""
    if (
        len(maybe_currency) == 6
        and maybe_currency[:3] in _CCY_CODES
        and maybe_currency[3:] in _CCY_CODES
    ):
        return maybe_currency + "=X"
    return maybe_currency


def symbol_to_currency(symbol: str) -> str:
    """Convert a currency pair to a YF currency symbol, prepending '=X'."""
    if (
        symbol.endswith("=X")
        and len(symbol) == 8
        and symbol[:3] in _CCY_CODES
        and symbol[3:6] in _CCY_CODES
    ):
        return symbol[:-2]
    return symbol


def _get_ticker(ticker: str) -> str:
    if (ticker_ := currency_to_symbol(ticker)) != ticker:
        logger.info("converting currency ticker %s to %s", ticker, ticker_)
        return ticker_
    return ticker


def sample_equity(
    ticker: str,
    start_date: str,
    end_date: str,
    interval: str = "1d",
    actions: bool = False,
    repair: bool = False,
):
    """sample one symbol from yahoo finance"""

    ticker = _get_ticker(ticker)

    logger.info(
        "querying yfinance ticker=%s start=%s end=%s interval=%s",
        ticker,
        start_date,
        end_date,
        interval,
    )

    if not end_date:
        raise ValueError("end_date must be defined")

    prices = yf.download(
        [ticker],
        start=pd.Timestamp(start_date) if start_date else None,
        end=pd.Timestamp(end_date),
        interval=interval,
        actions=actions,
        repair=repair,
        multi_level_index=False,
        threads=False,
        group_by="ticker",
        auto_adjust=True,
        prepost=True,
        progress=False,
        keepna=True,
    )
    if prices is None:
        raise ProviderDataError(f"Yahoo Finance returned no data for {ticker=}")

    if not isinstance(prices.index, pd.DatetimeIndex):
        prices.index = pd.to_datetime(prices.index, utc=True)
    if not prices.index.tz:
        prices = prices.tz_localize("utc")

    return (prices.tz_convert("utc"),)


@symbols.lib_provider(pricelib="{raw_price_lib}")
def create_universe(
    pricelib: Library,
    instruments: pd.DataFrame,
    end_date: pd.Timestamp,
    start_date: pd.Timestamp,
):
    """
    Create one arctic symbol for each OHLCV prices from yahoo finance.
    Each symbol contains all tickers defined for the universe.
    """

    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    def get_data(symbol: str):
        return pricelib.read(symbol, date_range=(start_date, end_date)).data

    symbols = pricelib.list_symbols()

    result = pd.concat(
        (
            (
                get_data(symbol)
                for symbol in instruments.index.to_list()
                if _get_ticker(symbol) in f"sample.{symbols}"
            )
        ),
        axis=1,
        keys=instruments.index.to_list(),
    ).reorder_levels([1, 0], axis=1)
    return (
        result["Open"],
        result["High"],
        result["Low"],
        result["Close"],
        result["Volume"],
    )


def adjust_fx_series(
    fx_series: pd.DataFrame,
    ref_ccy: str,
    add_self: bool = False,
    add_cent: bool = False,
) -> pd.DataFrame:
    """
    Adjust fx_series columns based on the reference currency.

    :param fx_series: DataFrame with XXXYYY columns where XXX and YYY are legs
    :param ref_ccy: Reference currency
    :return: Adjusted fx_series with renamed columns and inverted rates where necessary
    """
    adjusted_fx = fx_series.copy()
    new_columns = {}

    for col in fx_series.columns:
        base, quote = col[:3], col[3:]
        if quote == ref_ccy:
            new_columns[col] = base
        elif base == ref_ccy:
            new_columns[col] = quote
            adjusted_fx[col] = 1.0 / adjusted_fx[col]
        else:
            raise ValueError(
                f"Column {col} does not match reference currency {ref_ccy}"
            )

    adjusted_fx.rename(columns=new_columns, inplace=True)

    if add_self:
        adjusted_fx[ref_ccy] = 1.0

    if add_cent and "GBP" in adjusted_fx.columns:
        adjusted_fx["GBp"] = adjusted_fx["GBP"] * 0.01
    if add_cent and "EUR" in fx_series.columns:
        adjusted_fx["c"] = adjusted_fx["EUR"] * 0.01

    return adjusted_fx.loc[:, ~adjusted_fx.columns.duplicated()]


def _align_series(
    series: pd.Series,
    other: pd.Series | float | int | None = None,
) -> tuple[pd.Series, pd.Series]:
    """
    align a series to another ahead of returns calculation

    :param series: reference series
    :param other: other series or value to align
    """

    if other is None:
        other = pd.Series(1.0, index=series.index, name=series.name)
    elif isinstance(other, (float, int)):
        other = pd.Series(other, index=series.index, name=series.name)
    elif isinstance(other, pd.Series):
        common_idx = series.index.intersection(other.index)
        series = series.reindex(common_idx).dropna()
        other = other.reindex(common_idx).dropna()
    else:
        raise TypeError(type(other))

    return series, other


def convert_prices_to_ccy(
    instruments: pd.DataFrame,
    prices: pd.DataFrame,
    fx_series: pd.DataFrame,
    currency: str,
) -> pd.DataFrame:
    """
    Convert prices to a common currency using fx_series.

    :param instruments: DataFrame with instrument symbols and their currencies
    :param prices: DataFrame with prices indexed by instrument symbols
    :param fx_series: DataFrame with FX rates indexed by currency pairs (e.g., 'EURUSD')
    :param currency: Target currency to convert prices to
    :return: List of DataFrames with prices converted to the target currency
    """

    symbols_ccys = instruments.currency.to_dict()
    ccy_syms_map: dict[str, list[str]] = {}
    for symbol, ccy in symbols_ccys.items():
        if ccy not in ccy_syms_map:
            ccy_syms_map[ccy] = []
        ccy_syms_map[ccy].append(symbol)

    converted = []
    for name, df in prices.items():
        df_fx = adjust_fx_series(
            fx_series[name], currency, add_self=True, add_cent=True
        )
        if set(symbols_ccys) != set(df.columns):
            raise ValueError(
                f"prices columns {set(df.columns)} do not match currency_map symbols {set(symbols_ccys)}"
            )
        if missing_ccys := set(ccy_syms_map).difference(set(df_fx.columns)):
            raise ValueError(f"fx_series columns miss currencies: {missing_ccys}")

        result = []
        for sym in df.columns:
            df_, fx_ = _align_series(df[sym].ffill(), df_fx[symbols_ccys[sym]])
            result.append(df_.mul(fx_).rename(df_.name))
        converted.append(pd.concat(result, axis=1)[df.columns])

    return converted



# Historical IG data may be downloaded from the following site.
# https://www.dukascopy.com/swiss/english/marketwatch/historical/
# https://forexsb.com/historical-forex-data
#
import argparse
import logging
import os
import re
from collections import defaultdict
from pathlib import Path
from typing import List, Optional

import pandas as pd
from arcticdb import Arctic

import tradingo.sampling as sampling

logger = logging.getLogger(__name__)

ASSET_MAPPING = {
    "USA500.IDXUSD": "IX.D.SPTRD.IFS.IP",
    "GAS.CMDUSD": "CC.D.NG.UMP.IP",
    "BRENT.CMDUSD": "CC.D.LCO.UMP.IP",
    "USTBOND.TRUSD": "IR.D.10YEAR100.FWM2.IP",
    "COCOA.CMDUSD": "CC.D.CC.UMP.IP",
}

MULTIPLIERS = {
    "USA500.IDXUSD": 1,
    "GAS.CMDUSD": 1000,
    "BRENT.CMDUSD": 100,
    "USTBOND.TRUSD": 100,
    "COCOA.CMDUSD": 1,
}


def cli_app():
    app = argparse.ArgumentParser("load-historical-data")

    app.add_argument("--path", type=Path, required=True, nargs="+")
    app.add_argument("--arctic-uri", required=True)
    app.add_argument("--dry-run", action="store_true")
    app.add_argument("--universe", required=True)
    app.add_argument("--provider", required=True)
    app.add_argument("--clean", action="store_true")
    app.add_argument("--end-date", type=pd.Timestamp, default=None)

    return app


# groups: symbol, frequency, field, start_date, end_date
FILE_REGEX = r"^([A-Z0-9\.]+)_Candlestick_([0-9]+_[MDHS])_(BID|ASK)_([0-9]{2}\.[0-9]{2}\.[0-9]{4})-([0-9]{2}\.[0-9]{2}\.[0-9]{4}).csv$"


def main():
    args = cli_app().parse_args()

    a = Arctic(args.arctic_uri)

    result = read_backfill(
        paths=args.path,
        arctic=a,
        dry_run=args.dry_run,
        universe=args.universe,
        provider=args.provider,
        clean=args.clean,
    )

    sampling.instruments.download_instruments(
        index_col=None,
        epics=ASSET_MAPPING.values(),
        universe=args.universe,
        arctic=a,
    )

    if args.dry_run:
        print(result)
    print("finished")


def read_backfill(
    paths: List[Path],
    end_date: Optional[pd.Timestamp] = None,
    **kwargs,
):
    data_files = defaultdict(list)

    for path in paths:
        for file in os.listdir(path):
            if match := re.match(FILE_REGEX, file):
                symbol, frequency, field, start_date, end_date = match.groups()

                # need to group by symbol

                data_files[(field.lower(), symbol)].append(path / file)

    # symbol, ohlc, symbol
    #
    #
    def read_file(f):
        logger.warning("Reading %s", f)
        out = pd.read_csv(
            f,
            index_col=0,
            date_format="%d.%m.%Y %H:%M:%S.%f",
        )
        out.index = pd.to_datetime(out.index)
        if out.index.name == "Local time":
            out.index = out.index.tz_convert("utc")
        elif out.index.name == "Gmt time":
            out.index = out.index.tz_localize("GMT").tz_convert("utc")
        else:
            raise ValueError(out.index.name)
        return out.rename_axis("DateTime")

    result = pd.concat(
        (
            pd.concat(read_file(f) for f in files).query("~index.duplicated()")
            for files in data_files.values()
        ),
        axis=1,
        keys=data_files.keys(),
    ).reorder_levels([0, 2, 1], axis=1)

    result = result.mul(pd.Series(MULTIPLIERS), level=2, axis=1)

    result.rename(columns=ASSET_MAPPING, inplace=True)

    if end_date:
        result = result[result.index <= end_date]

    return (
        (result["bid"]["Open"], ("bid", "open")),
        (result["bid"]["High"], ("bid", "high")),
        (result["bid"]["Low"], ("bid", "low")),
        (result["bid"]["Close"], ("bid", "close")),
        (result["ask"]["Open"], ("ask", "open")),
        (result["ask"]["High"], ("ask", "high")),
        (result["ask"]["Low"], ("ask", "low")),
        (result["ask"]["Close"], ("ask", "close")),
        (((result["ask"]["Open"] + result["bid"]["Open"]) / 2), ("mid", "open")),
        (((result["ask"]["High"] + result["bid"]["High"]) / 2), ("mid", "high")),
        (((result["ask"]["Low"] + result["bid"]["Low"]) / 2), ("mid", "low")),
        (((result["ask"]["Close"] + result["bid"]["Close"]) / 2), ("mid", "close")),
    )



import logging
from typing import Optional

import numpy as np
import pandas as pd

from . import _backtest  # type: ignore

logger = logging.getLogger(__name__)


BACKTEST_FIELDS = (
    "unrealised_pnl",
    "realised_pnl",
    "total_pnl",
    "net_investment",
    "net_position",
    "net_exposure",
    "avg_open_price",
    "stop_trade",
)


def backtest(
    *,
    portfolio: pd.DataFrame,
    bid_close: pd.DataFrame,
    ask_close: pd.DataFrame,
    dividends: Optional[pd.DataFrame] = None,
    stop_limit: Optional[pd.DataFrame] = None,
    stop_loss: Optional[pd.DataFrame] = None,
    price_ffill_limit: int = 0,
) -> tuple[pd.DataFrame]:
    bid_close = bid_close.groupby(bid_close.index.date).ffill(limit=price_ffill_limit)
    ask_close = ask_close.groupby(bid_close.index.date).ffill(limit=price_ffill_limit)

    mid_close = (bid_close + ask_close) / 2
    bid_close, ask_close = (
        bid_close.reindex(mid_close.index),
        ask_close.reindex(mid_close.index),
    )

    # restrict position changes to where we have a price.
    portfolio = (
        portfolio.ffill()
        .reindex(mid_close.index, method="ffill")
        .where(mid_close.notna(), np.nan)
        .ffill()
        .fillna(0.0)
    )

    # portfolio = portfolio.reindex(mid_close.index, method="ffill").fillna(0.0)
    trades = portfolio.diff()
    trades.iloc[0] = portfolio.iloc[0]

    if dividends is None:
        dividends = pd.DataFrame(
            0,
            index=mid_close.index,
            columns=mid_close.columns,
        )
    if stop_limit is None:
        stop_limit = pd.DataFrame(
            np.nan,
            index=mid_close.index,
            columns=mid_close.columns,
        )
    if stop_loss is None:
        stop_loss = pd.DataFrame(
            np.nan,
            index=mid_close.index,
            columns=mid_close.columns,
        )

    def compute_backtest(inst_trades: pd.Series):
        logger.info("Computing backtest for ticker=%s", inst_trades.name)
        inst_mids = mid_close[inst_trades.name].ffill().dropna()
        inst_asks = ask_close[inst_trades.name].reindex(inst_mids.index, method="ffill")
        inst_bids = bid_close[inst_trades.name].reindex(inst_mids.index, method="ffill")
        inst_limit = stop_limit[inst_trades.name].reindex(inst_mids.index)
        inst_loss = stop_loss[inst_trades.name].reindex(inst_mids.index)

        inst_dividends = (
            dividends[inst_trades.name].reindex(mid_close.index).fillna(0.0)
        )

        return pd.DataFrame(
            data=_backtest.compute_backtest(
                inst_trades.reindex(inst_mids.index)
                .fillna(0.0)
                .to_numpy()
                .astype("float32"),
                inst_bids.to_numpy().astype("float32"),
                inst_asks.to_numpy().astype("float32"),
                inst_limit.to_numpy().astype("float32"),
                inst_loss.to_numpy().astype("float32"),
                inst_dividends.to_numpy().astype("float32"),
            ),
            index=inst_mids.index,
            columns=BACKTEST_FIELDS,
        )

    backtest = pd.concat(
        (compute_backtest(data) for _, data in trades.items()),
        keys=trades.columns,
        axis=1,
    ).reorder_levels([1, 0], axis=1)

    net_exposure = (backtest["net_position"] * mid_close).ffill().sum(axis=1)
    gross_exposure = (backtest["net_position"].abs() * mid_close).ffill().sum(axis=1)

    summary = (
        backtest[
            [
                "net_investment",
                "unrealised_pnl",
                "realised_pnl",
                "total_pnl",
            ]
        ]
        .ffill()
        .transpose()
        .groupby(level=0)
        .sum()
        .transpose()
    )
    summary["net_exposure"] = net_exposure
    summary["gross_exposure"] = gross_exposure

    backtest["unrealised_pnl"] = backtest["unrealised_pnl"].where(
        backtest.net_position.ne(0.0), np.nan
    )
    backtest_fields = (backtest.loc[:, f] for f in BACKTEST_FIELDS if f != "date")

    return (summary, *backtest_fields)



import pytest

from tradingo.dag import DAG


def test_dag_configuration():
    nodes = {
        "raw_prices": {
            "MSFT.sample": {
                "name": "MSFT.sample",
                "function": "tradingo.sampling.ig.sample_instrument",
                "depends_on": [],
                "symbols_in": [],
                "symbols_out": [
                    "ig-trading/{symbol}.mid",
                    "ig-trading/{symbol}.bid",
                    "ig-trading/{symbol}.ask",
                ],
                "params": {
                    "symbol": "MSFT",
                },
            },
            "AAPL.sample": {
                "name": "AAPL.sample",
                "function": "tradingo.sampling.ig.sample_instrument",
                "symbols_in": [],
                "depends_on": [],
                "symbols_out": [
                    "ig-trading/mid",
                    "ig-trading/bid",
                    "ig-trading/ask",
                ],
                "params": {
                    "symbol": "AAPL",
                },
            },
        },
        "prices": {
            "universe.sample": {
                "function": "tradingo.sampling.ig.sample_universe",
                "depends_on": ["AAPL.sample", "MSFT.sample"],
                "symbols_in": [
                    "ig-trading/AAPL.mid",
                    "ig-trading/AAPL.bid",
                    "ig-trading/AAPL.ask",
                    "ig-trading/MSFT.mid",
                    "ig-trading/MSFT.bid",
                    "ig-trading/MSFT.ask",
                ],
                "symbols_out": [
                    "prices/{universe}.mid.open",
                    "prices/{universe}.mid.high",
                    "prices/{universe}.mid.low",
                    "prices/{universe}.mid.close",
                    "prices/{universe}.bid.open",
                    "prices/{universe}.bid.high",
                    "prices/{universe}.bid.low",
                    "prices/{universe}.bid.close",
                    "prices/{universe}.ask.open",
                    "prices/{universe}.ask.high",
                    "prices/{universe}.ask.low",
                    "prices/{universe}.ask.close",
                ],
                "params": {},
            },
        },
        "signals": {
            "signal.trend": {
                "function": "tradingo.signals.trend",
                "symbols_out": ["signals/{universe}.trend"],
                "depends_on": ["universe.sample"],
                "symbols_in": ["prices/{universe}.mid.close"],
                "params": {
                    "prices": "prices/mid.close",
                    "library": "signals",
                    "field": "trend",
                    "universe": "ig-trading",
                },
            },
        },
    }

    dag = DAG.from_config(nodes)

    assert dag.get_symbols() == [
        "ig-trading/{symbol}.mid",
        "ig-trading/{symbol}.bid",
        "ig-trading/{symbol}.ask",
        "ig-trading/mid",
        "ig-trading/bid",
        "ig-trading/ask",
        "prices/{universe}.mid.open",
        "prices/{universe}.mid.high",
        "prices/{universe}.mid.low",
        "prices/{universe}.mid.close",
        "prices/{universe}.bid.open",
        "prices/{universe}.bid.high",
        "prices/{universe}.bid.low",
        "prices/{universe}.bid.close",
        "prices/{universe}.ask.open",
        "prices/{universe}.ask.high",
        "prices/{universe}.ask.low",
        "prices/{universe}.ask.close",
        "signals/{universe}.trend",
    ]


if __name__ == "__main__":
    pytest.main([__file__])



import pathlib

import pandas as pd
from arcticdb.version_store.library import Library

from tradingo import symbols


def load_backfill(start_date, end_date, symbol, interval, data_dir: pathlib.Path):

    return (
        (
            pd.read_csv(
                f"{data_dir}/{symbol}{interval}.csv",
                sep="\t",
                parse_dates=True,
                date_format="%Y-%m-%d %H:%M",
                index_col=0,
            )
            .set_axis(["open", "high", "low", "close", "volume"], axis=1)
            .rename_axis("timestamp")
            .tz_localize("utc")
            .loc[start_date:end_date]
        ),
    )


@symbols.lib_provider(pricelib="{raw_price_lib}")
def create_universe(
    pricelib: Library,
    symbols: list[str],
    end_date: pd.Timestamp,
    start_date: pd.Timestamp,
):
    start_date = pd.Timestamp(start_date)
    end_date = pd.Timestamp(end_date)

    result = pd.concat(
        (
            (
                pricelib.read(symbol, date_range=(start_date, end_date)).data
                for symbol in symbols
            )
        ),
        axis=1,
        keys=(s.rsplit(".", maxsplit=1)[0] for s in symbols),
    ).reorder_levels([1, 0], axis=1)
    return (
        result["open"],
        result["high"],
        result["low"],
        result["close"],
        result["volume"],
    )



"""DAG generation logic."""

from __future__ import annotations

import importlib
import json
import re
from enum import Enum
from pathlib import Path
from typing import Any, Callable, Iterable, Optional

from . import symbols
from .config import ConfigLoadError


class TaskState(Enum):
    """State of a task."""

    PENDING = "PENDING"
    FAILED = "FAILED"
    SUCCESS = "SUCCESS"


class Task:
    """
    Tradingo base task.
    It wraps an arbitrary function via configs, handling args, kwargs,
    input, output and other dependencies.
    """

    def __init__(
        self,
        name,
        function,
        task_args,
        task_kwargs,
        symbols_out,
        symbols_in,
        load_args,
        publish_args,
        dependencies: Iterable[str] = (),
    ) -> None:
        self._function = function
        self.name = name
        self.task_args = task_args
        self.task_kwargs = task_kwargs
        self._dependencies = list(dependencies)
        self._resolved_dependencies: list[Task] = []
        self.state = TaskState.PENDING
        self.symbols_out = symbols_out
        self.symbols_in = symbols_in
        self.load_args = load_args
        self.publish_args = publish_args

    def __repr__(self) -> str:
        return (
            f"Task(name='{self.name}', function='{self._function}',"
            f" task_args={self.task_args},"
            f" task_kwargs={self.task_kwargs},"
            f" dependcies={self._dependencies}, "
            ")"
        )

    @property
    def function(self) -> Callable:
        """task function"""
        module, function_name = self._function.rsplit(".", maxsplit=1)
        function = getattr(importlib.import_module(module), function_name)

        if self.symbols_out:
            function = symbols.symbol_publisher(
                *self.symbols_out,
                **self.publish_args,
            )(function)
        if self.symbols_in:
            function = symbols.symbol_provider(
                **self.symbols_in,
                **self.load_args,
            )(function)

        return function

    @staticmethod
    def prepare_kwargs(
        task_kwargs: dict[str, Any], global_kwargs: dict[str, Any]
    ) -> dict[str, Any]:
        """prepare kwargs to be passed to the function"""
        task_kwargs.update(global_kwargs)
        return task_kwargs

    def run(
        self,
        *args,
        run_dependencies: bool | int = False,
        skip_deps: re.Pattern | None = None,
        force_rerun=False,
        **kwargs,
    ) -> None:
        """run this task. optinally run also dependency tasks"""
        if run_dependencies:
            if isinstance(run_dependencies, int):
                run_dependencies -= 1
            for dependency in self.dependencies:
                if skip_deps and skip_deps.match(dependency.name):
                    continue
                dependency.run(
                    *args,
                    run_dependencies=run_dependencies,
                    force_rerun=force_rerun,
                    **kwargs,
                )

        state = self.state
        try:
            task_kwargs = self.prepare_kwargs(self.task_kwargs, kwargs)
            if self.state == TaskState.PENDING or force_rerun:
                state = TaskState.FAILED
                print(f"Running {self}")
                self.function(*self.task_args, *args, **task_kwargs)
                self.state = state = TaskState.SUCCESS
        finally:
            self.state = state

    def add_dependencies(self, *dependency) -> None:
        """add dependencies to this task"""
        self._dependencies.extend(dependency)

    def resolve_dependencies(self, tasks: dict[str, Task]) -> None:
        """resolve dependencies to this task"""
        self._resolved_dependencies.extend(
            tasks[dep_name] for dep_name in self._dependencies
        )

    @property
    def dependencies(self) -> list[Task]:
        """this tasks' other dependency tasks"""
        return self._resolved_dependencies

    @property
    def dependency_names(self) -> list[str]:
        """this tasks' other dependency task names"""
        return self._dependencies


def collect_task_configs(
    config, _tasks: Optional[dict[str, Any]] = None
) -> dict[str, dict]:
    """gather all task specifications from a config, accounting for dependencies."""
    tasks = _tasks or {}

    for key, value in config.items():
        if isinstance(value, dict) and "depends_on" in value:
            # its a task, collect it
            tasks[key] = value
        elif isinstance(value, dict):
            # its a set of tasks collect them
            tasks.update(collect_task_configs(value, tasks))

    return tasks


class DAG(dict[str, Task]):
    """Tradingo DAG"""

    @classmethod
    def from_config(cls, config: dict[str, Any]) -> DAG:
        """create a DAG from one config."""

        task_configs = collect_task_configs(config)

        tasks: dict[str, Task] = {}

        for task_name, task_config in task_configs.items():
            if not task_config.get("enabled", True):
                continue
            params = task_config["params"]
            try:
                tasks[task_name] = Task(
                    name=task_name,
                    function=task_config["function"],
                    task_args=(),
                    task_kwargs=params,
                    dependencies=task_config["depends_on"],
                    symbols_in=task_config.get("symbols_in", {}),
                    load_args=task_config.get("load_args", {}),
                    publish_args=task_config.get("publish_args", {}),
                    symbols_out=task_config.get("symbols_out", []),
                )
            except KeyError as ex:
                raise ConfigLoadError(
                    f"{task_name} is missing setting {ex.args[0]}"
                ) from ex

        for task_name, task in tasks.items():
            try:
                task.resolve_dependencies(tasks)
            except KeyError as ex:
                raise ConfigLoadError(
                    f"Missing task in dag '{ex.args[0]}' for '{task_name}'"
                ) from ex

        return cls(tasks)

    @property
    def _state_filepath(self) -> Path:
        return Path.home() / ".tradingo/dag-state.json"

    def print(self, include_dependencies: bool = True) -> None:
        """print the tasks of this DAG."""
        for task_name, task in self.items():
            print(f"{task_name}:")
            if not include_dependencies:
                continue
            if task:
                for dep in task.dependency_names:
                    print(f"  - {dep}")
            else:
                print("  No dependencies")
            print()

    def get_symbols(self) -> list[Task]:
        """list all symbols produced by this DAG."""
        return [
            task
            for subl in (task.symbols_out for task in self.values())
            for task in subl
        ]

    def run(self, task_name, skip_deps: re.Pattern[str], **kwargs) -> None:
        """run a specific task of this DAG."""
        if task_name not in self:
            raise ValueError(f"{task_name} is not a task in the DAG.")
        return self[task_name].run(**kwargs, skip_deps=skip_deps)

    def update_state(self) -> None:
        """update the local json file which keeps the DAG state."""

        if not self._state_filepath.exists():
            return

        dag_state = json.loads(self._state_filepath.read_text())

        for k, v in dag_state.items():
            if k not in self:
                continue
            state = TaskState[v]

            self[k].state = state if state == TaskState.SUCCESS else TaskState.PENDING

    def serialise_state(self) -> None:
        """write the DAG state into a local json file."""

        self._state_filepath.parent.mkdir(parents=True, exist_ok=True)
        self._state_filepath.write_text(
            json.dumps({k: v.state.value for k, v in self.items()}, indent=2)
        )
